/*
 * This is jack-keyboard 1.4, a virtual keyboard for JACK MIDI.
 * For questions and comments, contact <trasz@pin.if.uz.zgora.pl>. 
 * This code is public domain, you can do with it whatever you want.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <sysexits.h>
#include <jack/jack.h>
#include <jack/midiport.h>
#include <jack/ringbuffer.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

struct Note {
	int currently_playing;	// 1 if note is being played
	int sustain;

	int x;		// distance between the left edge of key and left edge of the window, in pixels
	int w;		// width of the key, in pixels
	int h;		// height of the key, in pixels
	int white;	// 1 if key is white; 0 otherwise;
};

#define NOTE_MAX 127

volatile struct Note notes[NOTE_MAX];

// table containing note values for the keys
int note_for_key[256];

#define VELOCITY_HIGH		100
#define VELOCITY_NORMAL		64
#define VELOCITY_LOW		32

int velocity=VELOCITY_NORMAL;

#define OUTPUT_PORT_NAME	"midi_out"
#define PROGRAM_NAME		"jack-keyboard"

jack_port_t *output_port;
int space_depressed=0;
int maybe_stop_sustained_notes=0;
int note_being_pressed_using_mouse=-1;
jack_client_t *jack_client=NULL;

#define NOTE_ON			0x90
#define NOTE_OFF		0x80
#define PROGRAM_CHANGE		0xC0
#define CONTROLLER		0xB0
#define RESET			0xFF
#define HOLD_PEDAL		64
#define ALL_SOUND_OFF		120
#define ALL_CONTROLLERS_OFF	121
#define ALL_NOTES_OFF		123
#define BANK_SELECT_COARSE	0

GtkWidget *window=NULL;

struct MidiMessage {
	int len;	// Length of MIDI message, in bytes.
	char data[3];
};

#define RINGBUFFER_SIZE		1024*sizeof(struct MidiMessage)

jack_ringbuffer_t *ringbuffer;

// Number of currently used program
int program=0;

// Number of currently selected bank
int bank=0;

double get_time(void)
{
	double seconds;
	int ret;
	struct timeval tv;
	
	ret=gettimeofday(&tv, NULL);
	if (ret) {
		perror("gettimeofday");
		exit(EX_OSERR);
	}

	seconds=tv.tv_sec+tv.tv_usec/1000000.0;

	return seconds;
}

int process_callback(jack_nframes_t nframes, void *arg)
{
	static double time_last_called=-1;
	double time_before_processing;
	double time_after_processing;

	time_before_processing=get_time();

	if (time_last_called<0)
		time_last_called=time_before_processing;

	if (time_before_processing - time_last_called > 0.1) {
		fprintf(stderr, "Had to wait %fs for JACK callback; possible note loss.\n",
			time_before_processing - time_last_called);
	}

	void* port_buffer = jack_port_get_buffer(output_port, nframes);

	jack_midi_clear_buffer(port_buffer, nframes);

	while (jack_ringbuffer_read_space(ringbuffer)) {
		struct MidiMessage ev;

		int read=jack_ringbuffer_read(ringbuffer, (char *)&ev, sizeof(ev));

		if (read!=sizeof(ev)) {
			fprintf(stderr, "Short read from ringbuffer, possible note loss.\n");
			continue;
		}

		unsigned char* buffer = jack_midi_event_reserve(port_buffer, 0, ev.len, nframes);
		if (buffer==NULL) {
			fprintf(stderr, "jack_midi_event_reserve failed, NOTE LOST.\n");
			continue;
		}

		if (ev.len>2)
			buffer[2]=ev.data[2];

		if (ev.len>1)
			buffer[1]=ev.data[1];

		buffer[0]=ev.data[0];
	}

	time_after_processing=get_time();

	if (time_after_processing - time_before_processing > 0.01) {
		fprintf(stderr, "Processing took %fs; scheduling problem?\n",
			time_after_processing - time_before_processing);
	}

	time_last_called=time_after_processing;

	return 0;
}

void queue_message(int b0, int b1, int b2)
{
	struct MidiMessage ev;

	if (b1==-1) {
		ev.len=1;
		ev.data[0]=b0;
	} else if (b2==-1) {
		ev.len=2;
		ev.data[0]=b0;
		ev.data[1]=b1;
	} else {
		ev.len=3;
		ev.data[0]=b0;
		ev.data[1]=b1;
		ev.data[2]=b2;
	}

	int written=jack_ringbuffer_write(ringbuffer, (char *)&ev, sizeof(ev));

	if (written!=sizeof(ev))
		fprintf(stderr, "Not enough space in the ringbuffer, NOTE LOST.\n");
}

void set_window_title(void)
{
	if (window==NULL)
		return;

	int i, off=0;
	char title[256];
	const char **connected_ports=jack_port_get_connections(output_port);

	off+=snprintf(title, sizeof(title)-off, "%s: bank %d, program %d",
		jack_get_client_name(jack_client), bank, program);

	if (connected_ports==NULL) {
		off+=snprintf(title+off, sizeof(title)-off, ", NOT CONNECTED");
	} else {
		off+=snprintf(title+off, sizeof(title)-off, ", connected to ");

		for (i=0; connected_ports[i]!=NULL; i++) {
			off+=snprintf(title+off, sizeof(title)-off, "%s%s",
			i==0 ? "" : ", ", connected_ports[i]);
		}
	}

	free(connected_ports);

	gtk_window_set_title(GTK_WINDOW(window), title);
}

int graph_order_callback(void *dummy)
{
	if (window!=NULL)
		set_window_title();

	return 0;
}

// XXX: two functions below are too long and somewhat similar in purpose.
void connect_to_next_input_port(void)
{
	const char **available_midi_ports=jack_get_ports(jack_client, NULL,
		JACK_DEFAULT_MIDI_TYPE, JackPortIsInput);

	if (available_midi_ports==NULL) {
		fprintf(stderr, "No midi input (listening) ports found.  Run some softsynth and try again.\n");

		return;
	}

	const char **connected_ports=jack_port_get_connections(output_port);
	const char *current=NULL, *next=NULL, *first_tried=NULL;
	int i;

	if (connected_ports)
		current=connected_ports[0];

	i=0;

	for (;;) {
		assert(available_midi_ports[i]!=NULL);

		if (current==NULL || !strcmp(available_midi_ports[i], current)) {

			if (available_midi_ports[i+1]!=NULL)
				next=available_midi_ports[i+1];
			else
				next=available_midi_ports[0];

			if (first_tried==next) {
				fprintf(stderr, "Cannot connect to any of the MIDI input ports.\n");
				goto out;
			}

			if (first_tried==NULL)
				first_tried=next;

			if (jack_port_get_connections(output_port)) {
				int ret=jack_port_disconnect(jack_client, output_port);
				if (ret) {
					fprintf(stderr, "Cannot disconnect MIDI port.\n");
					goto out;
				}
			}

			int ret=jack_connect(jack_client, jack_port_name(output_port), next);
			if (ret) {
				fprintf(stderr, "Cannot connect to %s.\n", next);
				current=next;
				continue;
			} else {

				fprintf(stderr, "Connected to %s.\n", next);
				set_window_title();
				goto out;
			}
		}

		i++;
	}

out:
	free(available_midi_ports);
	free(connected_ports);
}

void connect_to_prev_input_port(void)
{
	const char **available_midi_ports=jack_get_ports(jack_client, NULL,
		JACK_DEFAULT_MIDI_TYPE, JackPortIsInput);

	if (available_midi_ports==NULL) {
		fprintf(stderr, "No midi input (listening) ports found.  Run some softsynth and try again.\n");

		return;
	}

	const char **connected_ports=jack_port_get_connections(output_port);
	const char *current=NULL, *prev=NULL, *first_tried=NULL;
	int i, max;

	if (connected_ports)
		current=connected_ports[0];

	// max is the highest possible index into available_midi_ports[], i.e. number of elements - 1.
	for (max=0; available_midi_ports[max+1]!=0; max++);

	i=max;

	for (;;) {
		assert(available_midi_ports[i]!=NULL);

		if (current==NULL || !strcmp(available_midi_ports[i], current)) {

			if (i > 0)
				prev=available_midi_ports[i-1];
			else
				prev=available_midi_ports[max];

			if (first_tried==prev) {
				fprintf(stderr, "Cannot connect to any of the MIDI input ports.\n");
				goto out;
			}

			if (first_tried==NULL)
				first_tried=prev;

			if (jack_port_get_connections(output_port)) {
				int ret=jack_port_disconnect(jack_client, output_port);
				if (ret) {
					fprintf(stderr, "Cannot disconnect MIDI port.\n");
					goto out;
				}
			}

			int ret=jack_connect(jack_client, jack_port_name(output_port), prev);
			if (ret) {
				fprintf(stderr, "Cannot connect to %s.\n", prev);
				current=prev;
				continue;
			} else {

				fprintf(stderr, "Connected to %s.\n", prev);
				set_window_title();
				goto out;
			}
		}

		i--;
	}

out:
	free(available_midi_ports);
	free(connected_ports);
}

void init_jack(int argc, char *argv[])
{
	jack_client=jack_client_open(PROGRAM_NAME, JackNoStartServer, NULL);

	if (jack_client==NULL) {
		fprintf(stderr, "JACK server not running?\n");
		exit(EX_UNAVAILABLE);
	}

	jack_set_process_callback(jack_client, process_callback, 0);
	jack_set_graph_order_callback(jack_client, graph_order_callback, 0);
	output_port=jack_port_register(jack_client, OUTPUT_PORT_NAME, JACK_DEFAULT_MIDI_TYPE,
		JackPortIsOutput, 0);

	if (jack_activate(jack_client)) {
		fprintf(stderr, "Cannot activate client.\n");
		exit(EX_UNAVAILABLE);
	}

	connect_to_next_input_port();
}

void draw_key(GtkWidget *widget, int x, int y, int w, int h, int is_white)
{
	GdkColor black={0, 0, 0, 0};
	GdkColor white={0, 65535, 65535, 65535};

	GdkGC *gc=widget->style->fg_gc[0];

	if (is_white)
		gdk_gc_set_rgb_fg_color(gc, &white);
	else
		gdk_gc_set_rgb_fg_color(gc, &black);

	gdk_draw_rectangle(widget->window, gc, TRUE, x, y, w, h);
	gdk_gc_set_rgb_fg_color(gc, &black);
	gdk_draw_rectangle(widget->window, gc, FALSE, x, y, w, h);
}

void draw_note(int note)
{
	int white=notes[note].white;

	if (notes[note].currently_playing)
		white=!white;

	draw_key(window, notes[note].x, 0, notes[note].w, notes[note].h, white);

	// We need to redraw black keys that partially obscure the white one.
	if (note < NOTE_MAX-2 && ! notes[note+1].white)
		draw_note(note+1);

	if (note >=0 && ! notes[note-1].white)
		draw_note(note-1);
}

void stop_unsustained_notes(void)
{
	int i;

	for (i=0; i < NOTE_MAX; i++) {
		if (notes[i].currently_playing && ! notes[i].sustain) {
			notes[i].currently_playing=0;
			queue_message(NOTE_OFF, i, velocity);
			draw_note(i);
		}
	}
}

void stop_sustained_notes(void)
{
	int i;

	for (i=0; i < NOTE_MAX; i++) {
		if (notes[i].currently_playing && notes[i].sustain) {
			notes[i].currently_playing=0;
			queue_message(NOTE_OFF, i, velocity);
			draw_note(i);
		}
	}
}

int press_key(int key)
{
	assert(key>=0);
	assert(key<NOTE_MAX);

	maybe_stop_sustained_notes=0;

	if (space_depressed)
		notes[key].sustain=1;
	else
		notes[key].sustain=0;

	if (notes[key].currently_playing)
		return 0;

	notes[key].currently_playing=1;

	queue_message(NOTE_ON, key, velocity);
	draw_note(key);

	return 1;
}

int release_key(int key)
{
	assert(key>=0);
	assert(key<NOTE_MAX);

	maybe_stop_sustained_notes=0;

	if (space_depressed)
		notes[key].sustain=1;
	else
		notes[key].sustain=0;

	if (! notes[key].currently_playing)
		return 0;

	if (notes[key].sustain)
		return 0;

	notes[key].currently_playing=0;

	queue_message(NOTE_OFF, key, velocity);
	draw_note(key);

	return 1;
}

void panic(void)
{
	int i;

	// These two have to be sent first, in case we have no room in the ringbuffer
	// for all these NOTE_OFF messages sent five lines below.
	queue_message(CONTROLLER, ALL_NOTES_OFF, 0);
	queue_message(CONTROLLER, ALL_SOUND_OFF, 0);

	for (i=0; i < NOTE_MAX; i++) {
		release_key(i);

		queue_message(NOTE_OFF, i, 0);
	}

	queue_message(CONTROLLER, HOLD_PEDAL, 0);
	queue_message(CONTROLLER, ALL_CONTROLLERS_OFF, 0);
	queue_message(CONTROLLER, ALL_NOTES_OFF, 0);
	queue_message(CONTROLLER, ALL_SOUND_OFF, 0);
	queue_message(RESET, -1, -1);
}

unsigned char char_for_keyval(guint kv)
{
	gchar *name;

	switch (kv) {
		case GDK_space: return ' ';
		case GDK_minus: case GDK_KP_Subtract: return '-';
		case GDK_plus: case GDK_KP_Add: return '+';
		case GDK_equal: return '=';
		case GDK_KP_Multiply: return '*';
		case GDK_KP_Divide: return '/';
		default:
			name=gdk_keyval_name(gdk_keyval_to_lower(kv));
			if (strlen(name)==1)
				return name[0];

			return 0;
	}
}

gint keyboard_event_handler(GtkWidget *widget, GdkEventKey *event, gpointer func_data)
{
	// 0 for octave #4 (as in C4)
	static int octave_shift=0;

	unsigned char key=char_for_keyval(event->keyval);

	window=widget;

	/*
	 * '+' key shifts octave up.
	 * '-' key shifts octave down.
	 */
	if (key=='+' || key=='=') {
		if (event->type==GDK_KEY_PRESS && octave_shift < 3) {
			stop_unsustained_notes();
			octave_shift++;
		}

		return TRUE;
	}

	if (key=='-') {
		if (event->type==GDK_KEY_PRESS  && octave_shift > -5) {
			stop_unsustained_notes();
			octave_shift--;
		}

		return TRUE;
	}

	/*
	 * Shift increases velocity, i.e. holding it while pressing note key will make the sound louder.
	 * Ctrl decreases velocity.
	 */
	if (event->keyval==GDK_Shift_L || event->keyval==GDK_Shift_R) {
		if (event->type==GDK_KEY_PRESS)
			velocity=VELOCITY_HIGH;
		else
			velocity=VELOCITY_NORMAL;

		return TRUE;

	}

	if (event->keyval==GDK_Control_L || event->keyval==GDK_Control_R) {
		if (event->type==GDK_KEY_PRESS)
			velocity=VELOCITY_LOW;
		else
			velocity=VELOCITY_NORMAL;

		return TRUE;
	}

	/*
	 * '*' character increases program number.
	 * '/' character decreases it.
	 */
	if (key=='*') {
		if (event->type==GDK_KEY_PRESS  && program<127) {
			program++;
			queue_message(PROGRAM_CHANGE, program, -1);
			set_window_title();
		}
		
		return TRUE;
	}

	if (key=='/') {
		if (event->type==GDK_KEY_PRESS && program > 0) {
			program--;
			queue_message(PROGRAM_CHANGE, program, -1);
			set_window_title();
		}

		return TRUE;
	}

	if (event->keyval==GDK_Home) {
		if (event->type==GDK_KEY_PRESS && bank < 127) {
			bank++;
			queue_message(CONTROLLER, BANK_SELECT_COARSE, bank);
			set_window_title();
		}

		return TRUE;
	}

	if (event->keyval==GDK_End) {
		if (event->type==GDK_KEY_PRESS && bank > 0) {
			bank--;
			queue_message(CONTROLLER, BANK_SELECT_COARSE, bank);
			set_window_title();
		}

		return TRUE;
	}

	if (event->keyval==GDK_Insert) {
		if (event->type==GDK_KEY_PRESS)
			connect_to_next_input_port();

		return TRUE;
	}

	if (event->keyval==GDK_Delete) {
		if (event->type==GDK_KEY_PRESS)
			connect_to_prev_input_port();

		return TRUE;
	}

	if (event->keyval==GDK_Escape) {
		if (event->type==GDK_KEY_PRESS)
			panic();

		return TRUE;
	}

	/*
	 * Spacebar works as a 'sustain' key.  Holding spacebar while releasing note will cause
	 * the note to continue.
	 * Pressing and releasing spacebar without pressing any note keys will make all the
	 * sustained notes end (it will send 'note off' midi messages for all the sustained notes).
	 */
	if (key==' ') {
		if (event->type==GDK_KEY_PRESS) {
			if (!space_depressed) {
				space_depressed=1;
				maybe_stop_sustained_notes=1;
			}

		} else {
			if (maybe_stop_sustained_notes)
				stop_sustained_notes();

			space_depressed=0;
		}

		return TRUE;
	}

	if (key==0) {
		fprintf(stderr, "Unknown keyval '%d'\n", event->keyval);
		return TRUE;
	}

	int i=note_for_key[key];
	if (i<0) {
		fprintf(stderr, "No binding for '%c'\n", key);
		return TRUE;
	}

	i+=octave_shift*12;

	assert(i>=0);
	assert(i<=NOTE_MAX);

	if (event->type==GDK_KEY_PRESS) {
		press_key(i);

	} else if (event->type==GDK_KEY_RELEASE) {
		release_key(i);
	}

	// state of one of the virtual keys has changed.
	return TRUE;
}

int get_note_for_xy(int x, int y)
{
	int width;
	int height;
	int note;

	gtk_window_get_size(GTK_WINDOW(window), &width, &height);

	if (y<=height/2) {
		for (note=0; note<NOTE_MAX-1; note++) {
			if (notes[note].white)
				continue;

			if (x>=notes[note].x && x<=notes[note].x+notes[note].w)
				return note;
		}
	}

	for (note=0; note<NOTE_MAX-1; note++) {
		if (! notes[note].white)
			continue;

		if (x>=notes[note].x && x<=notes[note].x+notes[note].w)
			return note;
	}

	return -1;
}

gboolean mouse_button_event_handler(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
	int x=event->x;
	int y=event->y;

	int note=get_note_for_xy(x, y);

	if (event->type==GDK_BUTTON_PRESS) {
		assert(note>=0);
		press_key(note);
		note_being_pressed_using_mouse=note;

	} else if (event->type==GDK_BUTTON_RELEASE) {
		if (note>=0) {
			release_key(note);
		} else {
			assert(note_being_pressed_using_mouse>=0);
			release_key(note_being_pressed_using_mouse);
		}

		note_being_pressed_using_mouse=-1;
			
	}

	return TRUE;
}

gboolean mouse_motion_event_handler(GtkWidget *widget, GdkEventMotion *event, gpointer user_data)
{
	if (note_being_pressed_using_mouse<0)
		return TRUE;

	int note=get_note_for_xy(event->x, event->y);
	
	if (note!=note_being_pressed_using_mouse && note >=0) {
		release_key(note_being_pressed_using_mouse);
		press_key(note);
		note_being_pressed_using_mouse=note;
	}

	return TRUE;
}

gboolean expose_event(GtkWidget *widget, GdkEventExpose *event, gpointer user_data)
{
	int width;
	int height;

	gtk_window_get_size(GTK_WINDOW(widget), &width, &height);

	// NOTE_MAX-1 is the number of notes.  There are 7 white keys per 12 notes in octave.
	int number_of_white_keys=(NOTE_MAX-1)*(7.0/12.0);

	int key_width=width/number_of_white_keys;
	int black_key_width=key_width*0.8;

	int note;
	int white_key=0;
	int white;

	for (note=0, white_key=0; note<NOTE_MAX-2; note++) {
		int note_in_octave=note%12;

		if (notes[note].currently_playing) {
			
			white=0;
		} else {
			white=1;
		}

		if (note_in_octave==1 || note_in_octave==3 || note_in_octave==6 ||
			note_in_octave==8 || note_in_octave==10) {

			// This note is black key; draw them later.
			continue;
		}
	
		// This note is white key

		notes[note].x=white_key*key_width;
		notes[note].w=key_width;
		notes[note].h=height;
		notes[note].white=1;
	
		draw_key(widget, notes[note].x, 0, notes[note].w, notes[note].h, white);

		white_key++;
	}

	// Draw black keys over the white ones.
	for (note=0, white_key=0; note<NOTE_MAX-2; note++) {
		int note_in_octave=note%12;

		if (notes[note].currently_playing) {
			
			white=0;
		} else {
			white=1;
		}

		if (note_in_octave==1 || note_in_octave==3 || note_in_octave==6 ||
			note_in_octave==8 || note_in_octave==10) {

			// This note is black key.
			white=!white;

			notes[note].x=white_key*key_width - black_key_width/2;
			notes[note].w=black_key_width;
			notes[note].h=height/2;
			notes[note].white=0;

			draw_key(widget, notes[note].x, 0, notes[note].w, notes[note].h, white);

			continue;
		}
		
		white_key++;
	}

	return TRUE;
}

void init_notes(void)
{
	int i;

	for (i=0; i<=255; i++)
		note_for_key[i]=-1;
		
	// Lower keyboard row - 'zxcvbnm'.
	note_for_key['z']=60; // C4
	note_for_key['s']=61;
	note_for_key['x']=62; 
	note_for_key['d']=63; 
	note_for_key['c']=64; 
	note_for_key['v']=65; 
	note_for_key['g']=66; 
	note_for_key['b']=67; 
	note_for_key['h']=68; 
	note_for_key['n']=69; 
	note_for_key['j']=70; 
	note_for_key['m']=71; 

	// Higher keyboard row, first octave - 'qwertyu'.
	note_for_key['q']=72; 
	note_for_key['2']=73; 
	note_for_key['w']=74; 
	note_for_key['3']=75; 
	note_for_key['e']=76; 
	note_for_key['r']=77; 
	note_for_key['5']=78; 
	note_for_key['t']=79; 
	note_for_key['6']=80; 
	note_for_key['y']=81; 
	note_for_key['7']=82; 
	note_for_key['u']=83; 

	// Higher keyboard row, the rest - 'iop'.
	note_for_key['i']=84; 
	note_for_key['9']=85; 
	note_for_key['o']=86; 
	note_for_key['0']=87; 
	note_for_key['p']=88; 
}

int main(int argc, char *argv[])
{
	memset((void *)notes, 0, sizeof(struct Note) * NOTE_MAX);

	ringbuffer=jack_ringbuffer_create(RINGBUFFER_SIZE);
	if (ringbuffer==NULL) {
		fprintf(stderr, "Cannot create ringbuffer.\n");
		exit(EX_SOFTWARE);
	}

	jack_ringbuffer_mlock(ringbuffer);

	init_jack(argc, argv);
	init_notes();

	gtk_init (&argc, &argv);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

	set_window_title();

	gtk_widget_set_app_paintable(window, TRUE);
	gtk_window_set_default_size(GTK_WINDOW(window), 730, 70);
	gtk_widget_add_events(window, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK);
	g_signal_connect(G_OBJECT(window), "destroy", G_CALLBACK(exit), NULL);
	g_signal_connect(G_OBJECT(window), "expose-event", G_CALLBACK(expose_event), NULL);
	g_signal_connect(G_OBJECT(window), "button-press-event", G_CALLBACK(mouse_button_event_handler), NULL);
	g_signal_connect(G_OBJECT(window), "button-release-event", G_CALLBACK(mouse_button_event_handler), NULL);
	g_signal_connect(G_OBJECT(window), "motion-notify-event", G_CALLBACK(mouse_motion_event_handler), NULL);
	g_signal_connect(G_OBJECT(window), "key-press-event", G_CALLBACK(keyboard_event_handler), NULL);
	g_signal_connect(G_OBJECT(window), "key-release-event", G_CALLBACK(keyboard_event_handler), NULL);
	gtk_widget_show(window);

	gtk_main();

	return 0;
}

